/*-
 * #%L
 * Smile CDR - CDR
 * %%
 * Copyright (C) 2016 - 2024 Smile CDR, Inc.
 * %%
 * All rights reserved.
 * #L%
 */
package com.example.endpoint.hl7v2.inbound.transaction;

import ca.cdr.api.fhir.interceptor.CdrHook;
import ca.cdr.api.fhir.interceptor.CdrPointcut;
import ca.cdr.api.pub.hl7v2.model.Hl7v2ToFhirConversionResultJson;
import ca.uhn.fhir.context.FhirContext;
import ca.uhn.fhir.interceptor.api.Interceptor;
import ca.uhn.fhir.util.BundleBuilder;
import org.hl7.fhir.r4.model.BooleanType;
import org.hl7.fhir.r4.model.Bundle;
import org.hl7.fhir.r4.model.CodeType;
import org.hl7.fhir.r4.model.Enumerations;
import org.hl7.fhir.r4.model.HumanName;
import org.hl7.fhir.r4.model.IdType;
import org.hl7.fhir.r4.model.Parameters;
import org.hl7.fhir.r4.model.Patient;
import org.hl7.fhir.r4.model.StringType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.IntStream;

@Interceptor
public class TestPreConvertInterceptorUsingBundleBuilder {
	private static final Logger ourLog = LoggerFactory.getLogger(TestPreConvertInterceptorUsingBundleBuilder.class);
	private static final String NL = System.getProperty("line.separator");

	@Autowired
	private FhirContext myFhirContext;

	/**
	 * Demonstrating using an {@code ITermConceptClientMappingSvc} to have access to terminology translation
	 */
	@CdrHook(CdrPointcut.HL7V2IN_PRE_HL7V2_TO_FHIR_MAPPING_PROCESSING)
	public Boolean hl7v2PreProcess(Hl7v2ToFhirConversionResultJson theConversionResultJson) {
		try {
			ourLog.info("Interceptor HL7V2IN_PRE_HL7V2_TO_FHIR_MAPPING_PROCESSING - started");

			BundleBuilder bundleBuilder = new BundleBuilder(myFhirContext);

			String[] firstNames = {"John", "Emma"};
			String[] lastNames = {"Doe", "Smith"};
			String[] genders = {"male", "female"};

			// Add 2 patients to list

			List<Patient> patientList = new ArrayList<>();
			IntStream.range(0, firstNames.length).forEach(i -> {
				Patient patient = getPatient(firstNames, lastNames, genders, i);
				patientList.add(patient);
			});

			// create bundle

			patientList.forEach(bundleBuilder::addTransactionCreateEntry);
			Bundle createBundle = (Bundle) bundleBuilder.getBundle();
			ourLog.debug(
					"{}create bundle: {}",
					NL,
					myFhirContext.newJsonParser().setPrettyPrint(true).encodeResourceToString(createBundle));

			// update bundle

			bundleBuilder = new BundleBuilder(myFhirContext);
			patientList.forEach(bundleBuilder::addTransactionUpdateEntry);
			Bundle updateBundle = (Bundle) bundleBuilder.getBundle();
			ourLog.debug(
					"{}Update bundle: {}",
					NL,
					myFhirContext.newJsonParser().setPrettyPrint(true).encodeResourceToString(updateBundle));

			// delete bundle

			bundleBuilder = new BundleBuilder(myFhirContext);
			patientList.forEach(bundleBuilder::addTransactionDeleteEntry);
			Bundle deleteBundle = (Bundle) bundleBuilder.getBundle();
			ourLog.debug(
					"{}Delete bundle: {}",
					NL,
					myFhirContext.newJsonParser().setPrettyPrint(true).encodeResourceToString(deleteBundle));

			// conditional create bundle

			final BundleBuilder conditionalCreateBuilder = new BundleBuilder(myFhirContext);
			patientList.forEach(p -> {
				String createCondition = "Patient?identifier="
						+ p.getIdElement().withResourceType("Patient").getValue();
				conditionalCreateBuilder.addTransactionCreateEntry(p).conditional(createCondition);
			});
			ourLog.debug(
					"{}Conditional create bundle: {}",
					NL,
					myFhirContext
							.newJsonParser()
							.setPrettyPrint(true)
							.encodeResourceToString(conditionalCreateBuilder.getBundle()));

			// conditional update bundle

			final BundleBuilder conditionalUpdateBuilder = new BundleBuilder(myFhirContext);
			patientList.forEach(p -> {
				String updateCondition = "Patient?identifier="
						+ p.getIdElement().withResourceType("Patient").getValue();
				conditionalUpdateBuilder.addTransactionUpdateEntry(p).conditional(updateCondition);
			});
			ourLog.debug(
					"{}Conditional update bundle: {}",
					NL,
					myFhirContext
							.newJsonParser()
							.setPrettyPrint(true)
							.encodeResourceToString(conditionalUpdateBuilder.getBundle()));

			// conditional delete bundle

			final BundleBuilder conditionalDeleteBuilder = new BundleBuilder(myFhirContext);
			patientList.forEach(p -> {
				String deleteCondition = "Patient?identifier=" + p.getId();
				conditionalDeleteBuilder.addTransactionDeleteConditionalEntry(deleteCondition);
			});
			ourLog.debug(
					"{}Conditional delete bundle: {}",
					NL,
					myFhirContext
							.newJsonParser()
							.setPrettyPrint(true)
							.encodeResourceToString(bundleBuilder.getBundle()));

			// create Patient if it doesn't exit or patch it if it does

			final BundleBuilder conditionalCreateAndPatchBuilder = new BundleBuilder(myFhirContext);

			Patient patient = patientList.get(0).setActive(true);
			IdType patientId = patient.getIdElement().withResourceType("Patient");
			String condition = "Patient?identifier=" + patientId.getValue();
			conditionalCreateAndPatchBuilder.addTransactionCreateEntry(patient).conditional(condition);

			Parameters patch = new Parameters();
			Parameters.ParametersParameterComponent op = patch.addParameter().setName("operation");
			op.addPart().setName("type").setValue(new CodeType("replace"));
			op.addPart().setName("path").setValue(new CodeType("Patient.active"));
			op.addPart().setName("value").setValue(new BooleanType(true));
			conditionalCreateAndPatchBuilder
					.addTransactionFhirPatchEntry(patientId, patch)
					.conditional(condition);

			ourLog.debug(
					"{}Conditional create or conditional patch bundle: {}",
					NL,
					myFhirContext
							.newJsonParser()
							.setPrettyPrint(true)
							.encodeResourceToString(conditionalCreateAndPatchBuilder.getBundle()));

		} finally {
			ourLog.info("Interceptor HL7V2IN_PRE_HL7V2_TO_FHIR_MAPPING_PROCESSING - ended");
		}
		return true; // continue processing
	}

	private static Patient getPatient(String[] firstNames, String[] lastNames, String[] genders, int i) {
		Patient patient = new Patient();
		String patientId = "Hl7V2-tx-util-test-patient-" + i;
		patient.setId(patientId);
		HumanName name = new HumanName();
		name.setFamily(lastNames[i]);
		StringType givenName = new StringType(firstNames[i]);
		name.setGiven(Collections.singletonList(givenName));
		patient.setName(Collections.singletonList(name));
		patient.setGender(Enumerations.AdministrativeGender.fromCode(genders[i % 2]));
		return patient;
	}
}
